import { mockProcessExit, mockProcessStderr, mockProcessStdout } from 'jest-mock-process';
import { makeProgram } from './binlib';
import { Command } from 'commander';
import mock from 'mock-fs';
import fs from 'fs';

let mockStdout: jest.SpyInstance;
let mockStderr: jest.SpyInstance;
let mockExit: jest.SpyInstance;
let program: Command;
let stdinContents = '';

function getContents() {
  return stdinContents;
}

jest.mock('./binutils.ts', () => ({
  stdinToString: jest.fn(() => getContents()),
}));

beforeEach(() => {
  mockStdout = mockProcessStdout();
  mockStderr = mockProcessStderr();
  program = makeProgram();
  program.exitOverride();
  mockExit = mockProcessExit();
  mock({
    '.env': 'ALA=makota',
    'ugly.env': '########\n\n SOME = value \n\n\n ANOTHER=  "value with space"',
    'first.env': 'SERVER_PORT=80\nSERVER_HOST=localhost',
    'second.env': 'SERVER_PORT=8080\nSERVER_PASSWORD=secret\nCLIENT_HOST=192.168.4.42',
    'third.env': 'CLIENT_PORT=3000',
    'broken.env': 'CLI ENT_PORT=3000',
    'read-only.env': mock.file({
      content: 'READ=only',
      mode: 0o444,
    }),
  });
  stdinContents = '';
});

afterEach(() => {
  mockStdout.mockRestore();
  mockStderr.mockRestore();
  mockExit.mockRestore();
  mock.restore();
});

describe('Update command', () => {
  it('creates empty output without params or stdin', () => {
    stdinContents = '';
    program.parse(['node', 'test']);
    expect(mockExit).toBeCalled();
    expect(mockStdout).toHaveBeenCalledWith('\n');
  });

  it('uses stdin input', () => {
    stdinContents = 'VARIABLE=VALUE';
    program.parse(['node', 'test']);
    expect(mockExit).toBeCalled();
    expect(mockStdout).toHaveBeenCalledWith('VARIABLE=VALUE\n');
  });

  it('appends empty stdin', () => {
    stdinContents = '';
    program.parse(['node', 'test', 'ANOTHER=ok']);
    expect(mockExit).toBeCalled();
    expect(mockStdout).toHaveBeenCalledWith('ANOTHER=ok\n');
  });

  it('reads from invalid stdin', () => {
    stdinContents = 'junks';
    program.parse(['node', 'test']);
    expect(mockExit).toHaveBeenCalledWith(1);
    expect(mockStderr).toBeCalledTimes(2);
    expect(mockStderr).toHaveBeenNthCalledWith(1, expect.stringMatching(/Updating failed, cannot parse source stdin/));
    expect(mockStderr).toHaveBeenNthCalledWith(2, expect.stringMatching(/PARSING FAILED/));
  });

  it('updates correct stdin', () => {
    stdinContents = 'TEST=works';
    program.parse(['node', 'test', 'TEST=new value']);
    expect(mockExit).toBeCalled();
    expect(mockStdout).toHaveBeenCalledWith('TEST="new value"\n');
  });

  it('appends correct stdin', () => {
    stdinContents = 'TEST=works';
    program.parse(['node', 'test', 'ANOTHER=ok']);
    expect(mockExit).toBeCalled();
    expect(mockStdout).toHaveBeenCalledWith('TEST=works\nANOTHER=ok\n');
  });

  it('writes result to file', () => {
    stdinContents = 'TEST=works';
    program.parse(['node', 'test', '-o', 'test.txt', 'TEST=new']);
    expect(mockExit).toBeCalled();
    expect(fs.readFileSync('test.txt').toString()).toEqual('TEST=new\n');
  });

  it('writes result to read-only file', () => {
    stdinContents = 'TEST=works';
    program.parse(['node', 'test', '-o', 'read-only.env', 'TEST=new']);
    expect(mockExit).toHaveBeenCalledWith(1);
    expect(mockStderr).toBeCalledTimes(2);
    expect(mockStderr).toHaveBeenNthCalledWith(1, expect.stringMatching(/Updating failed - other error/));
    expect(mockStderr).toHaveBeenNthCalledWith(2, expect.stringMatching(/EACCES/));
    expect(mockStderr).toHaveBeenNthCalledWith(2, expect.stringMatching(/read-only.env/));
  });

  it('reads from correct file', () => {
    program.parse(['node', 'test', '-i', '.env', '--', 'ALA=ma psa']);
    expect(mockExit).toBeCalled();
    expect(mockStdout).toHaveBeenCalledWith('ALA="ma psa"\n');
  });

  it('beautify input file', () => {
    program.parse(['node', 'test', '-b', '-i', 'ugly.env']);
    expect(mockExit).toBeCalled();
    expect(mockStdout).toHaveBeenCalledWith('########\n\nSOME=value\n\nANOTHER="value with space"\n');
  });

  it('reading non-existing file', () => {
    program.parse(['node', 'test', '-b', '-i', 'i-am-not-here.env']);
    expect(mockExit).toHaveBeenCalledWith(1);
    expect(mockStderr).toBeCalledTimes(2);
    expect(mockStderr).toHaveBeenNthCalledWith(1, expect.stringMatching(/Updating failed, cannot read file: i-am-not-here.env/));
    expect(mockStderr).toHaveBeenNthCalledWith(2, expect.stringMatching(/ENOENT/));
    expect(mockStderr).toHaveBeenNthCalledWith(2, expect.stringMatching(/i-am-not-here.env/));
  });

  it('modify input file', () => {
    fs.writeFileSync('inplace.env', '########\n\n SOME = value \n\n\n ANOTHER=  "value with space"'),
      program.parse(['node', 'test', '-b', '-m', '-i', 'inplace.env']);
    expect(mockExit).toBeCalled();
    expect(mockStdout).toHaveBeenCalledTimes(0);
    expect(fs.readFileSync('inplace.env').toString()).toEqual('########\n\nSOME=value\n\nANOTHER="value with space"\n');
  });

  it('do not modify input file if invalid', () => {
    const data = '########\n\n IN VALID = value \n\n\n ANOTHER=  "value with space"';
    fs.writeFileSync('invalid.env', data), program.parse(['node', 'test', '-b', '-m', '-i', 'invalid.env']);
    expect(mockExit).toHaveBeenCalledWith(1);
    expect(mockStdout).toHaveBeenCalledTimes(0);
    expect(fs.readFileSync('invalid.env').toString()).toEqual(data);
    expect(mockStderr).toBeCalledTimes(4);
    // normally, exit will finish
    expect(mockStderr).toHaveBeenNthCalledWith(1, expect.stringMatching(/Updating failed, cannot parse source file invalid.env/));
    expect(mockStderr).toHaveBeenNthCalledWith(2, expect.stringMatching(/PARSING FAILED/));
    expect(mockStderr).toHaveBeenNthCalledWith(3, expect.stringMatching(/Overwriting invalid source file not possible: invalid.env/));
    expect(mockStderr).toHaveBeenNthCalledWith(4, expect.stringMatching(/PARSING FAILED/));
  });

  it('merge files', () => {
    program.parse(['node', 'test', '-i', 'first.env', 'second.env', 'third.env']);
    expect(mockExit).toBeCalled();
    expect(mockStdout).toHaveBeenCalledWith('SERVER_PORT=8080\nSERVER_HOST=localhost\nSERVER_PASSWORD=secret\nCLIENT_HOST=192.168.4.42\nCLIENT_PORT=3000\n');
  });

  it('merge files without update', () => {
    program.parse(['node', 'test', '-a', '-i', 'first.env', 'second.env', 'third.env']);
    expect(mockExit).toBeCalled();
    expect(mockStdout).toHaveBeenCalledWith('SERVER_PORT=80\nSERVER_HOST=localhost\nSERVER_PASSWORD=secret\nCLIENT_HOST=192.168.4.42\nCLIENT_PORT=3000\n');
  });

  it('merge files with smart append', () => {
    program.parse(['node', 'test', '-p', '-i', 'first.env', 'second.env', 'third.env']);
    expect(mockExit).toBeCalled();
    expect(mockStdout).toHaveBeenCalledWith('CLIENT_HOST=192.168.4.42\nCLIENT_PORT=3000\nSERVER_PASSWORD=secret\nSERVER_PORT=8080\nSERVER_HOST=localhost\n');
  });

  it('merge files with smart append without update', () => {
    program.parse(['node', 'test', '-p', '-a', '-i', 'first.env', 'second.env', 'third.env']);
    expect(mockExit).toBeCalled();
    expect(mockStdout).toHaveBeenCalledWith('CLIENT_HOST=192.168.4.42\nCLIENT_PORT=3000\nSERVER_PASSWORD=secret\nSERVER_PORT=80\nSERVER_HOST=localhost\n');
  });

  it('merge files and set from parameter', () => {
    program.parse(['node', 'test', '-i', 'first.env', 'second.env', 'third.env', '--', 'SERVER_PASSWORD=updated value']);
    expect(mockExit).toBeCalled();
    expect(mockStdout).toHaveBeenCalledWith(
      'SERVER_PORT=8080\nSERVER_HOST=localhost\nSERVER_PASSWORD="updated value"\nCLIENT_HOST=192.168.4.42\nCLIENT_PORT=3000\n',
    );
  });

  it('merge files and set from parameter without update', () => {
    program.parse(['node', 'test', '-a', '-i', 'first.env', 'second.env', 'third.env', '--', 'SERVER_PASSWORD=tried updated value', 'NEW_VALUE=append']);
    expect(mockExit).toBeCalled();
    expect(mockStdout).toHaveBeenCalledWith(
      'SERVER_PORT=80\nSERVER_HOST=localhost\nSERVER_PASSWORD=secret\nCLIENT_HOST=192.168.4.42\nCLIENT_PORT=3000\nNEW_VALUE=append\n',
    );
  });

  it('merge files and set from parameter and invalid parameter', () => {
    program.parse(['node', 'test', '-i', 'first.env', 'second.env', 'third.env', '--', 'SERVER_PASSWORD=updated value', 'INVA LID', 'LAST_ARGUMENT=correct']);
    expect(mockExit).toHaveBeenCalledWith(1);
    expect(mockStderr).toBeCalledTimes(2);
    expect(mockStderr).toHaveBeenNthCalledWith(1, expect.stringMatching(/Updating failed, cannot parse params passed/));
    expect(mockStderr).toHaveBeenNthCalledWith(2, expect.stringMatching(/Invalid argument/));
    expect(mockStderr).toHaveBeenNthCalledWith(2, expect.stringMatching(/INVA LID/));
  });

  it('merge files and set from parameter and invalid parameter', () => {
    program.parse([
      'node',
      'test',
      '-i',
      'first.env',
      'second.env',
      'third.env',
      '--',
      'SERVER_PASSWORD=updated value',
      'INVA-LID=wrong',
      'LAST_ARGUMENT=correct',
    ]);
    expect(mockExit).toHaveBeenCalledWith(1);
    expect(mockStderr).toBeCalledTimes(2);
    expect(mockStderr).toHaveBeenNthCalledWith(1, expect.stringMatching(/Updating failed, invalid param/));
    expect(mockStderr).toHaveBeenNthCalledWith(2, expect.stringMatching(/PARSING FAILED/));
    expect(mockStderr).toHaveBeenNthCalledWith(2, expect.stringMatching(/INVA-LID/));
  });

  it('merge files with invalid file name', () => {
    program.parse(['node', 'test', '-i', 'first.env', 'invalid.env', 'third.env']);
    expect(mockExit).toHaveBeenCalledWith(1);
    expect(mockStderr).toBeCalledTimes(2);
    expect(mockStderr).toHaveBeenNthCalledWith(1, expect.stringMatching(/Updating failed, cannot read file: invalid.env/));
    expect(mockStderr).toHaveBeenNthCalledWith(2, expect.stringMatching(/ENOENT/));
    expect(mockStderr).toHaveBeenNthCalledWith(2, expect.stringMatching(/invalid.env/));
  });

  it('merge files with broken file contents', () => {
    program.parse(['node', 'test', '-i', 'first.env', 'second.env', 'broken.env']);
    expect(mockExit).toHaveBeenCalledWith(1);
    expect(mockStderr).toBeCalledTimes(2);
    expect(mockStderr).toHaveBeenNthCalledWith(1, expect.stringMatching(/Updating failed, cannot parse file: broken.env/));
    expect(mockStderr).toHaveBeenNthCalledWith(2, expect.stringMatching(/PARSING FAILED/));
  });
});

describe('Update command resiliency', () => {
  it('reads from invalid stdin', () => {
    stdinContents = 'junks';
    program.parse(['node', 'test', '-r']);
    expect(mockExit).toBeCalled();
    expect(mockStderr).toHaveBeenCalledTimes(0);
    expect(mockStdout).toHaveBeenCalledWith('\n');
  });

  it('reading non-existing file', () => {
    program.parse(['node', 'test', '-r', '-b', '-i', 'i-am-not-here.env']);
    expect(mockExit).toBeCalled();
    expect(mockStderr).toHaveBeenCalledTimes(0);
    expect(mockStdout).toHaveBeenCalledWith('\n');
  });

  it('writes result to read-only file', () => {
    stdinContents = 'TEST=works';
    program.parse(['node', 'test', '-r', '-o', 'read-only.env', 'TEST=new']);
    expect(mockExit).toBeCalled();
    expect(mockStderr).toHaveBeenCalledTimes(0);
  });

  it('do not modify input file if invalid', () => {
    const data = '########\n\n IN VALID = value \n\n\n ANOTHER=  "value with space"';
    fs.writeFileSync('invalid.env', data), program.parse(['node', 'test', '-r', '-b', '-m', '-i', 'invalid.env']);
    expect(mockExit).toHaveBeenCalledWith(1);
    expect(mockStdout).toHaveBeenCalledTimes(0);
    expect(fs.readFileSync('invalid.env').toString()).toEqual(data);
    expect(mockStderr).toBeCalledTimes(2);
    expect(mockStderr).toHaveBeenNthCalledWith(1, expect.stringMatching(/Overwriting invalid source file not possible: invalid.env/));
    expect(mockStderr).toHaveBeenNthCalledWith(2, expect.stringMatching(/PARSING FAILED/));
  });

  it('merge files with invalid file name', () => {
    program.parse(['node', 'test', '-r', '-i', 'first.env', 'invalid.env', 'third.env']);
    expect(mockExit).toBeCalled();
    expect(mockStderr).toHaveBeenCalledTimes(0);
    expect(mockStdout).toHaveBeenCalledWith('SERVER_PORT=80\nSERVER_HOST=localhost\nCLIENT_PORT=3000\n');
  });

  it('merge files with broken file contents', () => {
    program.parse(['node', 'test', '-r', '-i', 'first.env', 'second.env', 'broken.env']);
    expect(mockExit).toBeCalled();
    expect(mockStderr).toHaveBeenCalledTimes(0);
    expect(mockStdout).toHaveBeenCalledWith('SERVER_PORT=8080\nSERVER_HOST=localhost\nSERVER_PASSWORD=secret\nCLIENT_HOST=192.168.4.42\n');
  });

  it('merge files and set from parameter and not parsable parameter', () => {
    program.parse([
      'node',
      'test',
      '-r',
      '-i',
      'first.env',
      'second.env',
      'third.env',
      '--',
      'SERVER_PASSWORD=updated value',
      'INVA-LID=par',
      'LAST_ARGUMENT=correct',
    ]);
    expect(mockExit).toBeCalled();
    expect(mockStderr).toHaveBeenCalledTimes(0);
    expect(mockStdout).toHaveBeenCalledWith(
      'SERVER_PORT=8080\nSERVER_HOST=localhost\nSERVER_PASSWORD="updated value"\nCLIENT_HOST=192.168.4.42\nCLIENT_PORT=3000\nLAST_ARGUMENT=correct\n',
    );
  });

  it('merge files and set from parameter and not parsable parameter', () => {
    program.parse([
      'node',
      'test',
      '-r',
      '-i',
      'first.env',
      'second.env',
      'third.env',
      '--',
      'SERVER_PASSWORD=updated value',
      'INVA LID',
      'LAST_ARGUMENT=correct',
    ]);
    expect(mockExit).toBeCalled();
    expect(mockStderr).toHaveBeenCalledTimes(0);
    expect(mockStdout).toHaveBeenCalledWith(
      'SERVER_PORT=8080\nSERVER_HOST=localhost\nSERVER_PASSWORD="updated value"\nCLIENT_HOST=192.168.4.42\nCLIENT_PORT=3000\nLAST_ARGUMENT=correct\n',
    );
  });
});

describe('Silent flag', () => {

  it('reads from invalid stdin', () => {
    stdinContents = 'junks';
    program.parse(['node', 'test', '-s']);
    expect(mockExit).toHaveBeenCalledWith(1);
    expect(mockStderr).toBeCalledTimes(0);
  });

  it('writes result to read-only file', () => {
    stdinContents = 'TEST=works';
    program.parse(['node', 'test', '-s', '-o', 'read-only.env', 'TEST=new']);
    expect(mockExit).toHaveBeenCalledWith(1);
    expect(mockStderr).toBeCalledTimes(0);
  });

  it('reading non-existing file', () => {
    program.parse(['node', 'test', '-s', '-b', '-i', 'i-am-not-here.env']);
    expect(mockExit).toHaveBeenCalledWith(1);
    expect(mockStderr).toBeCalledTimes(0);
  });

  it('do not modify input file if invalid', () => {
    const data = '########\n\n IN VALID = value \n\n\n ANOTHER=  "value with space"';
    fs.writeFileSync('invalid.env', data), program.parse(['node', 'test', '-s', '-b', '-m', '-i', 'invalid.env']);
    expect(mockExit).toHaveBeenCalledWith(1);
    expect(mockStderr).toBeCalledTimes(0);
  });

  it('merge files and set from parameter and invalid parameter', () => {
    program.parse([
      'node',
      'test',
      '-s',
      '-i',
      'first.env',
      'second.env',
      'third.env',
      '--',
      'SERVER_PASSWORD=updated value',
      'INVA LID',
      'LAST_ARGUMENT=correct',
    ]);
    expect(mockExit).toHaveBeenCalledWith(1);
    expect(mockExit).toHaveBeenCalledWith(1);
    expect(mockStderr).toBeCalledTimes(0);
  });

  it('merge files and set from parameter and invalid parameter', () => {
    program.parse([
      'node',
      'test',
      '-s',
      '-i',
      'first.env',
      'second.env',
      'third.env',
      '--',
      'SERVER_PASSWORD=updated value',
      'INVA-LID=wrong',
      'LAST_ARGUMENT=correct',
    ]);
    expect(mockExit).toHaveBeenCalledWith(1);
    expect(mockStderr).toBeCalledTimes(0);
  });

  it('merge files with invalid file name', () => {
    program.parse(['node', 'test', '-s', '-i', 'first.env', 'invalid.env', 'third.env']);
    expect(mockExit).toHaveBeenCalledWith(1);
    expect(mockStderr).toBeCalledTimes(0);
  });

  it('merge files with broken file contents', () => {
    program.parse(['node', 'test', '-s', '-i', 'first.env', 'second.env', 'broken.env']);
    expect(mockExit).toHaveBeenCalledWith(1);
    expect(mockStderr).toBeCalledTimes(0);
  });

  it('do not modify input file if invalid', () => {
    const data = '########\n\n IN VALID = value \n\n\n ANOTHER=  "value with space"';
    fs.writeFileSync('invalid.env', data), program.parse(['node', 'test', '-s', '-r', '-b', '-m', '-i', 'invalid.env']);
    expect(mockExit).toHaveBeenCalledWith(1);
    expect(mockStderr).toBeCalledTimes(0);
  });
});

describe('Update command without appending', () => {
  it('do not appends empty stdin', () => {
    stdinContents = '';
    program.parse(['node', 'test', '-u', 'ANOTHER=ok']);
    expect(mockExit).toBeCalled();
    expect(mockStdout).toHaveBeenCalledWith('\n');
  });

  it('do not appends correct stdin', () => {
    stdinContents = 'TEST=works';
    program.parse(['node', 'test', '-u', 'ANOTHER=ok']);
    expect(mockExit).toBeCalled();
    expect(mockStdout).toHaveBeenCalledWith('TEST=works\n');
  });

  it('merge files', () => {
    program.parse(['node', 'test', '-u', '-i', 'first.env', 'second.env', 'third.env']);
    expect(mockExit).toBeCalled();
    expect(mockStderr).toBeCalledTimes(0);
    expect(mockStdout).toHaveBeenCalledWith('SERVER_PORT=8080\nSERVER_HOST=localhost\n');
  });

  it('merge files and set from parameter and not append new', () => {
    program.parse(['node', 'test', '-u', '-i', 'first.env', 'second.env', 'third.env', '--', 'SERVER_PORT=443', 'LAST_ARGUMENT=correct']);
    expect(mockExit).toBeCalled();
    expect(mockStderr).toBeCalledTimes(0);
    expect(mockStdout).toHaveBeenCalledWith('SERVER_PORT=443\nSERVER_HOST=localhost\n');
  });
});
